/************************************************************************************/
/* Copyright (c) 2003, Outside Echo Ltd. & Hewlett-Packard.                         */
/* All rights reserved.                                                             */  
/*                                                                                  */
/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met:                    */
/*                                                                                  */
/*    * Redistributions of source code must retain the above copyright notice, this */
/*      list of conditions and the following disclaimer.                            */
/*    * Redistributions in binary form must reproduce the above copyright notice,   */
/*      this list of conditions and the following disclaimer in the documentation   */
/*      and/or other materials provided with the distribution.                      */
/*    * Neither the name of Outside Echo Ltd. & Hewlett-Packard                     */
/*      nor the names of its contributors may be used to endorse or promote products*/ 
/*      derived from this software without specific prior written permission.       */
/*                                                                                  */ 
/*                                                                                  */
/*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS NAMELY     */
/*   OUTSIDE ECHO LTD AND HEWLETT-PACKARD "AS IS" AND ANY                           */
/*   EXPRESS  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/*   WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/*   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE    */
/*   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL     */
/*   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR     */
/*   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER     */
/*   CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  */
/*   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  */
/*   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  DAMAGE.          */
/*                                                                                  */
/************************************************************************************/
/************************************************************************************/
/*                                                                                  */
/* This is a greedy implementation of the Set Covering Problem                      */
/* Algorithm can be found in Cormen's Book pg:974                                   */
/*                                                                                  */
/* The program takes into consideration weight.                                     */
/*                                                                                  */
/* The program assumes that complete coverage is possible and hence weight          */
/* is equal to 1/frequency.                                                         */
/*                                                                                  */
/************************************************************************************/
/*                                                                                  */
/* AUTHOR: Partha Pratim Talukdar, Hewlett-Packard Labs (India)                     */
/* Email: (ptpartha@india.hp.com)                                                   */
/* MODIFIED: Jan 7, 2004                                                            */
/*                                                                                  */
/************************************************************************************/

#include "basic.h"

using namespace std;


/* Instance of this class represents the list of units to be
 * covered. The class provides methods to access and modify the
 * list of units.
 */
class UnitList
{
   private:

      /* Information of each unit to be covered is stored in this
       * structure.
       */
      struct Unit
      {
	    char   unitStr[MAXUNITLEN];   /* Text representation of the unit */
	    float  unitWeight;            /* Weight/cost of the unit         */

	    /* Indicates whether the unit has been covered. */
	    bool   isCovered;         

	    /* Usage of this is like the sticky bit in page replacement      
	     * algorithms in OS. This bit is set if the unit is covered  
	     * in the last selected sentence. It is reset in the ResetSticky
	     * method.
	     */
	    bool   stickyFlag;       
	    
	    struct Unit *next;
      };

      typedef struct Unit Unit;

      Unit    *unitHead;
      int     totalUnits;              /* Total number of units in the unit list.
					* Maximum number is MAXROW.              
					*/   
      
      bool    reduceCompleteWeight;    /* This flag determines the type of weight 
					* calculation to be used                  
					*/

   public:

      UnitList()  /* Constructor */
      {
	 unitHead = NULL;
	 totalUnits = 0;
      }

      /* This method sets a flag which determines the type of   
       * unit weight adjustment to be used. The options are      
       * 1) To ignore the unit weight once it is covered.
       *    The flag is set when this type is used.              
       * 2) To decrement its weight gradually.                   
       *    The flag is reset when this type is used.            
       */
      void SetCalculationType(bool type)
      {
	 reduceCompleteWeight = type;
      }

      /* Mark units as covered which are present in a particular 
       * column (sentence).
       */
      void MarkCovered(char *selectedSentence)
      {
	 Unit *unitIterPtr;

	 unitIterPtr = unitHead;

	 while(unitIterPtr != NULL)
	 {
	    if(strstr(selectedSentence, unitIterPtr->unitStr) != NULL)
	    {		  
	       unitIterPtr->stickyFlag = true; 
	       unitIterPtr->isCovered = true;
	    }

	   unitIterPtr = unitIterPtr->next;
	 }

	 return;
      }

      /* Returns the count of uncovered units. 
       */
      int UncoveredCount()  
      {
	 int count;
	 Unit *tmpPtr;

	 count = 0;

	 tmpPtr = unitHead;

	 while(tmpPtr != NULL)
	 {
	    if(tmpPtr->isCovered == false)
	       ++count;

	    tmpPtr = tmpPtr->next;
	 }

	 return(count);
      }
      
      /* Returns the weight of a sentence. This method is called 
       * while loading sentences into memory.
       */
      float GetWeight(string str)
      {
 	 Unit  *unitIterPtr;
	 float tmpWeight;

	 unitIterPtr = unitHead;
	 tmpWeight = 0;

	 while(unitIterPtr != NULL)
	 {
	    if((unitIterPtr->isCovered == false) && strstr(str.c_str(), unitIterPtr->unitStr) != NULL)
	    {
	       tmpWeight += unitIterPtr->unitWeight;
	    }

	    unitIterPtr = unitIterPtr->next;
	 }

	 return(tmpWeight);
      }

      /* Returns the total weight of units which are present in the 
       * currently selected sentence. This is used by the MaxCoverSentence 
       * method which selects sentences on an iterative basis.                                            
       */
      float CalculateWeight(const char *str)
      {
 	 Unit  *unitIterPtr;
	 float tmpWeight;

	 unitIterPtr = unitHead;
	 tmpWeight = 0;

	 while(unitIterPtr != NULL)
	 {
	    /* A set sticjy flag indicates that the unit has been covered in the most recently
	     * selected sentence.
	     */
	    if((strstr(str, unitIterPtr->unitStr) != NULL) && (unitIterPtr->stickyFlag == true))
	    {
	       tmpWeight += unitIterPtr->unitWeight;
	    }
	   
	    unitIterPtr = unitIterPtr->next;
	 }
	    
	 return(tmpWeight);
      }      

      /* Resets the sticky bit and adjusts the weight of the units.
       */
      void ResetSticky()
      {
	 Unit *unitIterPtr;

	 unitIterPtr = unitHead;

	 while(unitIterPtr != NULL)
	 {
	    if(unitIterPtr->stickyFlag == true)
	    {
	       if(unitIterPtr->unitWeight > 0)
	       {
		  if(reduceCompleteWeight)
		  {			
		     unitIterPtr->unitWeight = 0;
		  }
		  else
		  {
		     unitIterPtr->unitWeight = (unitIterPtr->unitWeight * 0.5); 
		  }
	       }
	       else
		  unitIterPtr->unitWeight = 0;
	    }

	    unitIterPtr->stickyFlag = false;
	    unitIterPtr = unitIterPtr->next;
	 }

	 return;
      }

      /* Returns the total no of units present.
       */
      int GetTotalUnits()   
      {
	 return totalUnits;
      } 
  
      /* Insert a unit and its weight into the unit list. 
       */
      void Insert(char str[], float weight)   
      {
	 Unit *tmpNode;

	 tmpNode = (Unit *) malloc(sizeof(Unit));

	 if(tmpNode == NULL)
	    Error::PrintAndExit("Memory Could Not Be Allocated: Program Terminated");

	 tmpNode->next = unitHead;

	 strcpy(tmpNode->unitStr, str);
	 tmpNode->unitWeight = weight;
	 tmpNode->isCovered = false;
	 tmpNode->stickyFlag = false;

	 unitHead = tmpNode;

	 /* printf("Inserted unit with weight : %s %f\n", tmpNode->unitStr, tmpNode->unitWeight); */

	 ++totalUnits;

	 return;
      }

      /* Prints the uncovered units and stores them in a file. 
       */
      void PrintUncoveredUnits()
      {
	 int count;
	 FILE *uncoverFP;
	 Unit *tmpPtr;

	 if((uncoverFP = fopen("uncovered_units","w")) == NULL)
	    Error::PrintAndExit("PrintSelectedSentence: File to print uncovered units could not be opened");

	 printf("\n\t ** Uncovered Units ** \n\n");

	 count = 0;

	 tmpPtr = unitHead;

	 while(tmpPtr != NULL)
	 {	 
	    if(tmpPtr->isCovered == false)
	    {
	       ++count;
	       printf("%s ", tmpPtr->unitStr);
	       fprintf(uncoverFP, "%s\n", tmpPtr->unitStr);
	    }

	    tmpPtr = tmpPtr->next;
	 }

	 printf("\n");

	 fclose(uncoverFP);
      }
};

/* Instance of this class represents the coverage matrix. The class 
 * provides various methods to operate on the mattrix. The rows of 
 * the matrix the units to becovered while the columns represent the
 * sentences in the corpus.
 */

class Matrix
{
   private:

      struct Sentence
      {
	    char  *senText;     /* Stores the sentence or its representation (e.g. phonetic) */
	    int    senLength;   /* Stores the length of the sentence.                        */
	    float  senWeight;   /* Stores the weight of a sentence.                          */
	    int    senNumber;   /* Stores the sentence offset.                               */
	    bool   isSelected;  /* Flag which indicates whether a sentence has been selected.*/
	    
	    struct Sentence *next;
      };

      typedef struct Sentence Sentence;

      Sentence *senHead;
      int      senCount;

   public:

      /* Constructor */
      Matrix()  
      {
	 senHead = NULL;
	 senCount = 0;
      }

      /* Returns number of columns currently used */
      int GetColumnCount()
      {
	 return(senCount);
      }

      /* This method takes in as argument the array representing
       * the set of already selected sentences and also the cost/weight 
       * of each unit so that the weight of each sentence can be 
       * calculated.
       *
       * Any future selection heuristic will change this method. 
       * Finds out the column which covers the maximum no of units.
       * Improvement may be done using MAX HEAP 
       *
       * This method returns the index of the selected column.
       * -1 is returned if none of the uncovered units can be 
       * covered by the remaining set of unselected sentences.
       */
      int MaxCoverSentence(UnitList objUnitList, int prevUncCount)  
      {
	 Sentence *selSenPtr, *senIterPtr;
	 float    maxCoverCount;  
    
	 senIterPtr = senHead;
	 selSenPtr  = senHead;
	 
	 while(senIterPtr != NULL)
	 {
	    if(senIterPtr->isSelected)
	    {
	       senIterPtr = senIterPtr->next;
	       continue;
	    }

	    if(senIterPtr->senWeight > selSenPtr->senWeight)
	    {
	       selSenPtr = senIterPtr;
	    }
	    else if(senIterPtr->senWeight == selSenPtr->senWeight)
	    {
	       /* If both weights are eqaual then select the sentence with 
		* smaller length. 
		*/
	       if(senIterPtr->senLength < selSenPtr->senLength)
		  selSenPtr = senIterPtr;
	    }

	    senIterPtr = senIterPtr->next;	  
	 }

	 // printf("Returned Weight: %2.10f\n", selSenPtr->senWeight);

	 objUnitList.MarkCovered(selSenPtr->senText);

	 if((selSenPtr->senWeight == 0.0) || (prevUncCount - objUnitList.UncoveredCount() == 0))
	 {
	    printf("Inside else if\n");

	    return -1;
	 }
	 else if(selSenPtr->senWeight > 0.0) 
	 {
	    selSenPtr->isSelected = true;
	    
    	    printf("Current Maximum Weight: %f %d\n", selSenPtr->senWeight, selSenPtr->senNumber);

	    Rearrange(objUnitList);
	 }	

	 return(1);
      }

      /* This method updates the matrix to reflect the changes caused 
       * by the current selection. 
       */
      void Rearrange(UnitList objUnitList)
      {
	 Sentence *senIterPtr;
	 float    currNodeWeight;

	 senIterPtr = senHead;

	 while(senIterPtr != NULL)
	 {
	   /** 
	    * Fixed on 25th March 2004.
	    * Keeps intact the weight of the selected sentence. 
	    */
	    if(senIterPtr->isSelected)
	    {
	       senIterPtr = senIterPtr->next;
	       continue;
	    }
	    
	    currNodeWeight = objUnitList.CalculateWeight(senIterPtr->senText);
	    
	    /* Adjust the weight of the sentences which are effected 
	     * by the current selection of a sentence and thereby 
	     * covering a few more units.
	     */
	    senIterPtr->senWeight = senIterPtr->senWeight - currNodeWeight;

	    if(senIterPtr->senWeight < 0)
	       senIterPtr->senWeight = 0.0;

	    senIterPtr = senIterPtr->next;
	 }

	 objUnitList.ResetSticky();
      }

      /* Updates the Matrix with the coverage of each sentence. 
       */
      void Insert(string sentenceStr, float sentenceWeight, int lineNumber)    
      {
	 Sentence *tmpSenNode;

	 tmpSenNode = (Sentence *) malloc(sizeof(Sentence));
	 tmpSenNode->senText = (char *) malloc(sentenceStr.length()+1);

	 if(tmpSenNode == NULL)
	    Error::PrintAndExit("Memory Could Not Be Allocated: Program Terminated");

	 strcpy(tmpSenNode->senText, sentenceStr.c_str());

	 tmpSenNode->senWeight = 0;
	 tmpSenNode->senWeight = sentenceWeight;

	 tmpSenNode->senLength = sentenceStr.length();
	 tmpSenNode->senNumber = lineNumber;
	 tmpSenNode->isSelected = false;

	 tmpSenNode->next = senHead;
	 senHead = tmpSenNode;

	 senCount++;

	 return;
      }

      /* This method prints out the selected sentences. 
       */
      void PrintSelectedSentence(char *corpusFileName)
      {
	 int  senCount, selCount;
	 FILE *prFP, *ofp;
	 Sentence *senIterPtr;

	 if((ofp = fopen("selected_sentences","w")) == NULL)
	    Error::PrintAndExit("PrintSelectedSentence : selected_sentences File Could Not be Opened ");

	 selCount = 0;
	 senIterPtr = senHead;

	 while(senIterPtr != NULL)
	 {
	    if(senIterPtr->isSelected == true)
	    {
	       ++selCount;
	       
	       OrthographicOutput(corpusFileName, senIterPtr->senNumber, senIterPtr->senWeight, ofp);
	    }
	    
	    senIterPtr = senIterPtr->next;
	 }

	 printf("\nTotal: %d Sentence(s) Selected\n\n", selCount);
//	 fprintf(ofp, "Total: %d\n", selCount);

	 fclose(ofp);
      }

       /* Input: corpusFileName: File from which sentences were ranked and 
       *                        marked for selection
       *        fixedOffset:    Offset of a particular selected sentence
       *                        in the file containing sentences.
       *        ofp:            Output File Pointer. 
       *
       * This method extracts the selected sentences and stores them in a file.
       * The sentences are extracted from the original_text file. This file is 
       * separate from the one from which the sentences were ranked and later 
       * on selected based on the weights. It is *assumed* that there is a one
       * to one correspondence between the sentences in the original_text file
       * and the file input by the user. It is hoped that the sentences in the
       * original_text file is in a more readable format and hence this method
       * is provided.
       *
       */
      void OrthographicOutput(char *corpusFileName, int fixedOffset, float sentenceWeight, FILE *ofp)
      {
	 int tmpOffset, firstCount, secCount;
	 int ch;
	 char tmpSentence[MAXSENLEN];
	 FILE *oriFP, *secFP; 

	 tmpOffset = 0;
	 firstCount =  0;
	 secCount = 0;

	 if((oriFP = fopen("original_text", "r")) == NULL)
	    Error::PrintAndExit("original_text File Not Found");
	
	 if((secFP = fopen(corpusFileName, "r")) == NULL)
	    Error::PrintAndExit("Input File Not Found");


	 while(tmpOffset < fixedOffset)
	 {
	    ++tmpOffset;

	    if(fgetc(secFP) == '\n')
	       ++firstCount;		
	 }

	 tmpOffset = 0;

	 if(firstCount)
	    while(secCount < firstCount)
	    {
	       tmpOffset++;

	       if(fgetc(oriFP) == '\n')
		  ++secCount;
	    }
	 
	 strcpy(tmpSentence, "");

	 fseek(oriFP, tmpOffset, SEEK_SET);
	 fscanf(oriFP,"%[^\n]\n", tmpSentence);

	 fprintf(ofp,"%f %s\n\n", sentenceWeight, tmpSentence);

	 fclose(oriFP);
	 fclose(secFP);

	 return;
      }	
};

/* Instance of this class does the core job of sentence seelection. 
 * A greedy approach is adopted while selecting sentences.
 */

class SelectSentence
{
   private:

      UnitList unitList;  /* Object where the units to be covered will be stored */ 
      Matrix   matrix;               
  
   public:

      /* This is the main method which calls other methods
       */
      void Run(char *corpusFileName, bool completeCoverage, bool reduceCompleteWeight)
      {
	 unitList.SetCalculationType(reduceCompleteWeight);

	 /* Load the units and their weights/costs 
	  */
	 UnitFileLoad(completeCoverage);

	 cout << "\nUnit File Loaded\n";

	 /* Load the sentences into memory and check the 
	  * also update the coverage matrix 
	  */
	 SentenceFileLoad(corpusFileName);

	 cout << "\nSentence File Loaded\n\n";

	 /* Start the sentence selection process 
	  */ 
	 StartSentenceSelection();

	 /* Print the Selected Sentences 
	  */
	 matrix.PrintSelectedSentence(corpusFileName);
      }

   private:

      /* This method selects sentences on an iterative basis */
      void StartSentenceSelection()
      {
	 int selectionResult;
	 int prevUncovered;

	 prevUncovered = 0;

	 /* Loop until all the units are covered or the corpus 
	  * is exhausted.                                
	  */
	 while((prevUncovered = unitList.UncoveredCount()) != 0)  
	 {
	    printf("\nCurrent Uncovered: %d\n", prevUncovered);

	    /* Select the sentence with maximum coverage of the 
	     * remaining uncovered units 
	     */
	    selectionResult = matrix.MaxCoverSentence(unitList, prevUncovered);

	    /* Check whether any of the remaining sentences can cover the 
	     * remaining uncovered units. If not, break out of this loop
	     * since more coverage is not possible.
	     */
	    if(selectionResult == -1)   
	    {
	       unitList.PrintUncoveredUnits();
	       break;
	    }	
	 }	
      }

      /* Loads the unit and their frequencies into memory and insert 
       * them in to the unit list along with their weights.
       */
      void UnitFileLoad(bool completeCoverage)   
      {
	 FILE *mapFP;
	 char unitStr[MAXUNITLEN];
	 int  fileUnitWeight;
	 float  unitWeight;

	 if((mapFP = fopen("units","r")) == NULL)
	    Error::PrintAndExit("Unit File Not Found");

	 while(fscanf(mapFP,"%s %d\n", unitStr, &fileUnitWeight) != EOF)
	 {
	    /* If complete coverage is possible the frequnecy = 1/freuqency
	     * else, it is kept intact
	     */
	    if(completeCoverage)
	    { 
	       /* If complete coverage is possible and the weight of the unit is greater than 0 
		* then the weight is made equal to inverse of the frequency so that the high 
		* frequency units get lower preference comapred to the lower frequency ones. Since
		* complete coverage is guaranteed so even the high frequency units will get covered
		* inspite of lowering their weights.
		*/
	       if(fileUnitWeight > 0) 
		  unitWeight = (1.0)/fileUnitWeight;
	       else
		  unitWeight = 0;
	    }
	    else
	    {
	       /* If complete coverage is not guaranteed then weight of the units is made equal 
		* to their frequencies.
		*/
	       unitWeight = fileUnitWeight;
	    }

	    unitList.Insert(unitStr, unitWeight);
	    strcpy(unitStr,"");
	 }

	 fclose(mapFP);

	 return;
      }

      /* Read in the sentences and build up the coverage matrix. 
       */
      void SentenceFileLoad(char *corpusFileName)  
      {
	 std::ifstream corFP;
	 string        sentence;
	 int           index, offset;

	 corFP.open(corpusFileName);

	 index = 1;
	 offset = 0;

	 while(!corFP.eof())
	 {
	    offset = corFP.tellg();
	    getline(corFP, sentence, '\n');

	    matrix.Insert(sentence, unitList.GetWeight(sentence), offset);
	    ++index;
	 }

	 corFP.close();
      }
};

/* The main method */

int main(int argc,char **argv)
{
   SelectSentence selectSentence;
   bool completeCoverage, reduceCompleteWeight;

   if(argc != 3)
      Error::PrintAndExit("Usage : ./text_sel <Corpus Filename> <Complete Coverage (Y/N)> \n\nComplete Coverage: Y -> Complete Coverage, N -> Complete Coverage Not Possible");

   if(strcmp(argv[2], "N") == 0)
      completeCoverage = false;
   else if(strcmp(argv[2], "Y") == 0)
      completeCoverage = true;
   else
      Error::PrintAndExit("Alternatives for the Second Option Are: \nY -> Complete Coverage \nN -> Complete Coverage Not Possible" );

   /* This flag controls the unit weight adjustment.
    * Once a unit is covered, its weight is made zero. 
    * The option is to reduce the weight of the unit 
    * stepwise. This feature is currently not incorporated.
    */
   reduceCompleteWeight = true;

   /* Pass on the filename of the corpus file containing 
    * the sentences, indication whether complete coverage 
    * is possible, how should the weight of the sentences
    * be chaged.
    */
   selectSentence.Run(argv[1], completeCoverage, reduceCompleteWeight); 
}

